@@ -81,6 +81,18 @@ static PyObject* setInPmap(PyObject* keySequence, Py_ssize_t keySize, PyObject*
   return newMap;
 }
 
+static PyObject* transform_fn = NULL;
+
+static PyObject* transform(PVector* self, PyObject* args) {
+  if(transform_fn == NULL) {
+    // transform to avoid circular import problems
+    transform_fn = PyObject_GetAttrString(PyImport_ImportModule("pyrsistent"), "_transform");
+  }
+
+  return PyObject_CallFunctionObjArgs(transform_fn, self, args, NULL);
+}
+
+
 // No access to internal members
 static PyMemberDef PVector_members[] = {
 	{NULL}  /* Sentinel */
@@ -142,7 +154,7 @@ static PVector* emptyNewPvec(void);
 static PVector* copyPVector(PVector *original);
 static void extendWithItem(PVector *newVec, PyObject *item);
 
-static PyObject *PVectorEvolver_pvector(PVectorEvolver *);
+static PyObject *PVectorEvolver_persistent(PVectorEvolver *);
 static int PVectorEvolver_set_item(PVectorEvolver *, PyObject*, PyObject*);
 
 static Py_ssize_t PVector_len(PVector *self) {
@@ -540,6 +552,8 @@ static PyObject* PVector_append(PVector *self, PyObject *obj);
 
 static PyObject* PVector_set_in(PVector *self, PyObject *obj);
 
+static PyObject* PVector_transform(PVector *self, PyObject *obj);
+
 static PyObject* PVector_set(PVector *self, PyObject *obj);
 
 static PyObject* PVector_mset(PVector *self, PyObject *args);
@@ -574,7 +588,8 @@ static PyMethodDef PVector_methods[] = {
 	{"set",         (PyCFunction)PVector_set, METH_VARARGS, "Inserts an element at the specified position"},
 	{"extend",      (PyCFunction)PVector_extend, METH_O|METH_COEXIST, "Extend"},
 	{"set_in",      (PyCFunction)PVector_set_in, METH_VARARGS, "Insert an element in a nested structure"},
-	{"index",       (PyCFunction)PVector_index, METH_VARARGS, "Return first index of value"},
+        {"transform",   (PyCFunction)PVector_transform, METH_VARARGS, "Apply one or more transformations"},
+        {"index",       (PyCFunction)PVector_index, METH_VARARGS, "Return first index of value"},
 	{"count",       (PyCFunction)PVector_count, METH_O, "Return number of occurrences of value"},
         {"__reduce__",  (PyCFunction)PVector_pickle_reduce, METH_NOARGS, "Pickle support method"},
         {"evolver",     (PyCFunction)PVector_evolver, METH_NOARGS, "Return new evolver for pvector"},
@@ -995,6 +1010,13 @@ static PyObject* PVector_set_in(PVector *self, PyObject *args) {
   }
 }
 
+
+static PyObject* PVector_transform(PVector *self, PyObject *obj) {
+  return transform(self, obj);
+}
+
+
+
 /*
  Steals a reference to the object that is inserted in the vector.
 */
@@ -1026,7 +1048,7 @@ static PyObject* PVector_mset(PVector *self, PyObject *args) {
     }
   }
 
-  PyObject* vector = PVectorEvolver_pvector(evolver);
+  PyObject* vector = PVectorEvolver_persistent(evolver);
   Py_DECREF(evolver);
   return vector;
 }
@@ -1223,7 +1245,7 @@ static void PVectorEvolver_dealloc(PVectorEvolver *);
 static PyObject *PVectorEvolver_append(PVectorEvolver *, PyObject *);
 static PyObject *PVectorEvolver_extend(PVectorEvolver *, PyObject *);
 static PyObject *PVectorEvolver_subscript(PVectorEvolver *, PyObject *);
-static PyObject *PVectorEvolver_pvector(PVectorEvolver *);
+static PyObject *PVectorEvolver_persistent(PVectorEvolver *);
 static Py_ssize_t PVectorEvolver_len(PVectorEvolver *);
 static PyObject *PVectorEvolver_is_dirty(PVectorEvolver *);
 static int PVectorEvolver_traverse(PVectorEvolver *self, visitproc visit, void *arg);
@@ -1238,7 +1260,7 @@ static PyMappingMethods PVectorEvolver_mapping_methods = {
 static PyMethodDef PVectorEvolver_methods[] = {
 	{"append",      (PyCFunction)PVectorEvolver_append, METH_O,       "Appends an element"},
 	{"extend",      (PyCFunction)PVectorEvolver_extend, METH_O|METH_COEXIST, "Extend"},
-        {"pvector",     (PyCFunction)PVectorEvolver_pvector, METH_NOARGS, "Create PVector from evolver"},
+        {"persistent",  (PyCFunction)PVectorEvolver_persistent, METH_NOARGS, "Create PVector from evolver"},
         {"is_dirty",    (PyCFunction)PVectorEvolver_is_dirty, METH_NOARGS, "Check if evolver contains modifications"},
         {NULL,              NULL}           /* sentinel */
 };
@@ -1455,16 +1477,15 @@ static int PVectorEvolver_set_item(PVectorEvolver *self, PyObject* item, PyObjec
   return -1;
 }
 
-static PyObject *PVectorEvolver_pvector(PVectorEvolver *self) {
+static PyObject *PVectorEvolver_persistent(PVectorEvolver *self) {
   PVector *resultVector;
-  if(self->newVector == self->originalVector) {
-    resultVector = self->newVector;
-  } else {
+  if(self->newVector != self->originalVector) {
     cleanVector(self->newVector);
-    resultVector = self->newVector;
     Py_DECREF(self->originalVector);
   }
 
+  resultVector = self->newVector;
+
   if(PyList_GET_SIZE(self->appendList)) {
     PVector *oldVector = resultVector;
     resultVector = (PVector*)PVector_extend(resultVector, self->appendList);
