@@ -4,6 +4,7 @@ from functools import wraps, reduce
 from itertools import chain, islice
 from numbers import Integral
 import sys
+import re
 
 import six
 
@@ -203,7 +204,8 @@ class PVector(object):
         return hash(self._totuple())
 
     class _Evolver(object):
-        __slots__ = ('_count', '_shift', '_root', '_tail', '_tail_offset', '_dirty_nodes', '_extra_tail', '_cached_leafs')
+        __slots__ = ('_count', '_shift', '_root', '_tail', '_tail_offset', '_dirty_nodes',
+                     '_extra_tail', '_cached_leafs', '_orig_vector')
 
         def __init__(self, v):
             self._reset(v)
@@ -229,6 +231,7 @@ class PVector(object):
             self._dirty_nodes = {}
             self._cached_leafs = {}
             self._extra_tail = []
+            self._orig_vector = v
 
         def append(self, element):
             self._extra_tail.append(element)
@@ -278,16 +281,19 @@ class PVector(object):
 
             return ret
 
-        def pvector(self):
-            v = PVector(self._count, self._shift, self._root, self._tail).extend(self._extra_tail)
-            self._reset(v)
-            return v
+        def persistent(self):
+            result = self._orig_vector
+            if self.is_dirty():
+                result = PVector(self._count, self._shift, self._root, self._tail).extend(self._extra_tail)
+                self._reset(result)
+
+            return result
 
         def __len__(self):
             return self._count + len(self._extra_tail)
 
         def is_dirty(self):
-            return self._dirty_nodes or self._extra_tail
+            return bool(self._dirty_nodes or self._extra_tail)
 
     def evolver(self):
         """
@@ -325,7 +331,7 @@ class PVector(object):
 
         The changes are kept in the evolver. An updated pvector can be created using the
         pvector() function on the evolver.
-        >>> v2 = e.pvector()
+        >>> v2 = e.persistent()
         >>> v2
         pvector([1, 22, 3, 4, 5, 6, 7, 8, 10])
 
@@ -352,7 +358,7 @@ class PVector(object):
         for i in range(0, len(args), 2):
             evolver[args[i]] = args[i+1]
 
-        return evolver.pvector()
+        return evolver.persistent()
 
     def set(self, i, val):
         """
@@ -562,6 +568,32 @@ class PVector(object):
         # Pickling support
         return _pvector, (self._tolist(),)
 
+    def transform(self, *transformations):
+        """
+        Transform arbitrarily complex combinations of PVectors and PMaps. A transformation
+        consists of two parts. One match expression that specifies which elements to transform
+        and one transformation function that performs the actual transformation.
+
+        >>> news_paper = freeze({'articles': [{'author': 'Sara', 'content': 'A short article'},
+        ...                                   {'author': 'Steve', 'content': 'A slightly longer article'}],
+        ...                      'weather': {'temperature': '11C', 'wind': '5m/s'}})
+        >>> short_news = news_paper.transform(['articles', ny, 'content'], lambda c: c[:25] + '...' if len(c) > 25 else c)
+        >>> very_short_news = news_paper.transform(['articles', ny, 'content'], lambda c: c[:15] + '...' if len(c) > 15 else c)
+        >>> very_short_news.articles[0].content
+        'A short article'
+        >>> very_short_news.articles[1].content
+        'A slightly long...'
+
+        # When nothing has been transformed the original data structure is kept
+        >>> short_news is news_paper
+        True
+        >>> very_short_news is news_paper
+        False
+        >>> very_short_news.articles[0] is news_paper.articles[0]
+        True
+        """
+        return _transform(self, transformations)
+
 
 Sequence.register(PVector)
 Hashable.register(PVector)
@@ -638,9 +670,11 @@ class PMap(object):
     """
     __slots__ = ('_size', '_buckets')
 
-    def __init__(self, size, buckets):
+    def __new__(cls, size, buckets):
+        self = super(PMap, cls).__new__(cls)
         self._size = size
         self._buckets = buckets
+        return self
 
     @staticmethod
     def _get_bucket(buckets, key):
@@ -718,7 +752,9 @@ class PMap(object):
 
     __eq__ = Mapping.__eq__
     __ne__ = Mapping.__ne__
-    __str__ = __repr__
+
+    def __str__(self):
+        return self.__repr__()
 
     def __hash__(self):
         # This hashing algorithm is probably not the speediest
@@ -740,19 +776,20 @@ class PMap(object):
         """
         evolver = self.evolver()
         evolver[key] = val
-        return evolver.pmap()
+        return evolver.persistent()
 
     def remove(self, key):
         """
         Return a new PMap without the element specified by key. Raises KeyError if the element
         is not present.
+
         >>> m1 = m(a=1, b=2)
         >>> m1.remove('a')
         pmap({'b': 2})
         """
         evolver = self.evolver()
         del evolver[key]
-        return evolver.pmap()
+        return evolver.persistent()
 
     def discard(self, key):
         """
@@ -802,7 +839,7 @@ class PMap(object):
             for key, value in map.items():
                 evolver[key] = update_fn(evolver[key], value) if key in evolver else value
 
-        return evolver.pmap()
+        return evolver.persistent()
 
     def set_in(self, keys, val):
         """
@@ -824,6 +861,35 @@ class PMap(object):
         # Pickling support
         return pmap, (dict(self),)
 
+    def transform(self, *transformations):
+        """
+        Transform arbitrarily complex combinations of PVectors and PMaps. A transformation
+        consists of two parts. One match expression that specifies which elements to transform
+        and one transformation function that performs the actual transformation.
+
+        >>> news_paper = freeze({'articles': [{'author': 'Sara', 'content': 'A short article'},
+        ...                                   {'author': 'Steve', 'content': 'A slightly longer article'}],
+        ...                      'weather': {'temperature': '11C', 'wind': '5m/s'}})
+        >>> short_news = news_paper.transform(['articles', ny, 'content'], lambda c: c[:25] + '...' if len(c) > 25 else c)
+        >>> very_short_news = news_paper.transform(['articles', ny, 'content'], lambda c: c[:15] + '...' if len(c) > 15 else c)
+        >>> very_short_news.articles[0].content
+        'A short article'
+        >>> very_short_news.articles[1].content
+        'A slightly long...'
+
+        # When nothing has been transformed the original data structure is kept
+        >>> short_news is news_paper
+        True
+        >>> very_short_news is news_paper
+        False
+        >>> very_short_news.articles[0] is news_paper.articles[0]
+        True
+        """
+        return _transform(self, transformations)
+
+    def copy(self):
+        return self
+
     class _Evolver(object):
         __slots__ = ('_buckets_evolver', '_size', '_original_pmap')
 
@@ -860,7 +926,7 @@ class PMap(object):
 
         def _reallocate(self, new_size):
             new_list = new_size * [None]
-            buckets = self._buckets_evolver.pvector()
+            buckets = self._buckets_evolver.persistent()
             for k, v in chain.from_iterable(x for x in buckets if x):
                 index = hash(k) % new_size
                 if new_list[index]:
@@ -873,9 +939,9 @@ class PMap(object):
         def is_dirty(self):
             return self._buckets_evolver.is_dirty()
 
-        def pmap(self):
+        def persistent(self):
             if self.is_dirty():
-                return PMap(self._size, self._buckets_evolver.pvector())
+                return PMap(self._size, self._buckets_evolver.persistent())
 
             return self._original_pmap
 
@@ -916,14 +982,14 @@ class PMap(object):
 
         The changes are kept in the evolver. An updated pmap can be created using the
         pmap() function on the evolver.
-        >>> m2 = e.pmap()
+        >>> m2 = e.persistent()
         >>> m2
         pmap({'c': 3, 'b': 2})
 
         The new pmap will share data with the original pmap in the same way that would have
         been done if only using operations on the pmap.
         """
-        return PMap._Evolver(self)
+        return self._Evolver(self)
 
 Mapping.register(PMap)
 Hashable.register(PMap)
@@ -1085,15 +1151,18 @@ class PSet(object):
         def is_dirty(self):
             return self._pmap_evolver.is_dirty()
 
-        def pset(self):
+        def persistent(self):
             if not self.is_dirty():
                 return  self._original_pset
 
-            return PSet(self._pmap_evolver.pmap())
+            return PSet(self._pmap_evolver.persistent())
 
         def __len__(self):
             return len(self._pmap_evolver)
 
+    def copy(self):
+        return self
+
     def evolver(self):
         """
         Create a new evolver for this pset. For a discussion on evolvers in general see the
@@ -1113,7 +1182,7 @@ class PSet(object):
 
         The changes are kept in the evolver. An updated pmap can be created using the
         pset() function on the evolver.
-        >>> s2 = e.pset()
+        >>> s2 = e.persistent()
         >>> s2
         pset([2, 3, 4])
 
@@ -1455,14 +1524,10 @@ class {class_name}(namedtuple('PClassBase', [{quoted_members}], verbose={verbose
     try:
         six.exec_(template, namespace)
     except SyntaxError as e:
-        raise e
         raise SyntaxError(e.message + ':\n' + template)
 
     return namespace[name]
 
-# Deprecated alias for pclass
-immutable = pclass
-
 ## Freeze & Thaw
 
 def freeze(o):
@@ -2194,3 +2259,337 @@ def dq(*elements):
     pdeque([1, 2, 3])
     """
     return pdeque(elements)
+
+
+###### PRecord ######
+
+class _PRecordMeta(type):
+    def __new__(mcs, name, bases, dct):
+        dct['_precord_fields'] = dict(sum([list(b.__dict__.get('_precord_fields', {}).items()) for b in bases], []))
+
+        for k, v in list(dct.items()):
+            if isinstance(v, _PRecordField):
+                dct['_precord_fields'][k] = v
+                del dct[k]
+
+        # Global invariants are inherited
+        dct['_precord_invariants'] = [dct['__invariant__']] if '__invariant__' in dct else []
+        dct['_precord_invariants'] += [b.__dict__['__invariant__'] for b in bases if '__invariant__' in b.__dict__]
+        if not all(callable(invariant) for invariant in dct['_precord_invariants']):
+            raise TypeError('Global invariants must be callable')
+
+        dct['_precord_mandatory_fields'] = \
+            set(name for name, field in dct['_precord_fields'].items() if field.mandatory)
+
+        dct['_precord_initial_values'] = \
+            dict((k, field.initial) for k, field in dct['_precord_fields'].items() if field.initial is not _PRECORD_NO_INITIAL)
+
+        dct['__slots__'] = ()
+
+        return super(_PRecordMeta, mcs).__new__(mcs, name, bases, dct)
+
+
+class InvariantException(Exception):
+    """
+    Exception raised from :py:class:`PRecord` when invariant tests fail or when a mandatory
+    field is missing.
+
+    Contains two fields of interest:
+    invariant_errors, a tuple of error data for the failing invariants
+    missing_fields, a tuple of strings specifying the missing names
+    """
+
+    def __init__(self, error_codes, missing_fields, *args, **kwargs):
+        self.invariant_errors = error_codes
+        self.missing_fields = missing_fields
+        super(InvariantException, self).__init__(*args, **kwargs)
+
+class _PRecordField(object):
+    __slots__ = ('type', 'invariant', 'initial', 'mandatory', 'factory', 'serializer')
+
+    def __init__(self, type, invariant, initial, mandatory, factory, serializer):
+        self.type = type
+        self.invariant = invariant
+        self.initial = initial
+        self.mandatory = mandatory
+        self.factory = factory
+        self.serializer = serializer
+
+_PRECORD_NO_TYPE = ()
+_PRECORD_NO_INVARIANT = lambda _: (True, None)
+_PRECORD_NO_FACTORY = lambda x: x
+_PRECORD_NO_INITIAL = object()
+_PRECORD_NO_SERIALIZER = lambda _, value: value
+
+def field(type=_PRECORD_NO_TYPE, invariant=_PRECORD_NO_INVARIANT, initial=_PRECORD_NO_INITIAL,
+          mandatory=False, factory=_PRECORD_NO_FACTORY, serializer=_PRECORD_NO_SERIALIZER):
+    """
+    Field specification factory for :py:class:`PRecord`.
+    type, a type or iterable with types that are allowed for this field
+    invariant, a function specifying an invariant that must hold for the field
+    initial, initial value of field if not specified when instantiating the record
+    mandatorty, boolean specifying if the field is mandatory or not
+    factory, factory function called when field is set.
+    serializer, function that returns a serialized versino of the field
+    """
+
+    types = set(type) if isinstance(type, Iterable) else set([type])
+
+    # If no factory is specified and the type is another PRecord use the factory method
+    # of that PRecord
+    if factory is _PRECORD_NO_FACTORY and len(types) == 1 and issubclass(tuple(types)[0], PRecord):
+        factory = tuple(types)[0].create
+
+    field = _PRecordField(type=types, invariant=invariant, initial=initial, mandatory=mandatory,
+                          factory=factory, serializer=serializer)
+
+    _check_field_parameters(field)
+
+    return field
+
+def _check_field_parameters(field):
+    for t in field.type:
+        if not isinstance(t, type):
+            raise TypeError('Type paramenter expected, not {0}'.format(type(t)))
+
+    if field.initial is not _PRECORD_NO_INITIAL and field.type and not any(isinstance(field.initial, t) for t in field.type):
+        raise TypeError('Initial has invalid type {0}'.format(type(t)))
+
+    if not callable(field.invariant):
+        raise TypeError('Invariant must be callable')
+
+    if not callable(field.factory):
+        raise TypeError('Factory must be callable')
+
+    if not callable(field.serializer):
+        raise TypeError('Serializer must be callable')
+
+def _restore_pickle(cls, data):
+    return cls.create(data)
+
+
+@six.add_metaclass(_PRecordMeta)
+class PRecord(PMap):
+    """
+    A PRecord is a PMap with a fixed set of specified fields. Records are declared as python classes inheriting
+    from PRecord. Because it is a PMap it has full support for all Mapping methods such as iteration and element
+    access using subscript notation.
+
+    More documentation and examples of PRecord usage is available at https://github.com/tobgu/pyrsistent
+    """
+    def __new__(cls, **kwargs):
+        # Hack total! If these two special attributes exist that means we can create
+        # ourselves. Otherwise we need to go through the Evolver to create the structures
+        # for us.
+        if '_precord_size' in kwargs and '_precord_buckets' in kwargs:
+            return super(PRecord, cls).__new__(cls, kwargs['_precord_size'], kwargs['_precord_buckets'])
+
+        initial_values = kwargs
+        if cls._precord_initial_values:
+            initial_values = dict(cls._precord_initial_values)
+            initial_values.update(kwargs)
+
+        e = _PRecordEvolver(cls, pmap())
+        for k, v in initial_values.items():
+            e[k] = v
+
+        return e.persistent()
+
+    def set(self, *args, **kwargs):
+        """
+        Set a field in the record. This set function differs slightly from that in the PMap
+        class. First of all it accepts key-value pairs. Second it accepts multiple key-value
+        pairs to perform one, atomic, update of multiple fields.
+        """
+
+        # The PRecord set() can accept kwargs since all fields that have been declared are
+        # valid python identifiers. Also allow multiple fields to be set in one operation.
+        if args:
+            return super(PRecord, self).set(args[0], args[1])
+
+        return self.update(kwargs)
+
+    def evolver(self):
+        """
+        Returns an evolver of this object.
+        """
+        return _PRecordEvolver(self.__class__, self)
+
+    def __repr__(self):
+        return "{0}({1})".format(self.__class__.__name__,
+                                 ', '.join('{0}={1}'.format(k, v) for k, v in self.items()))
+
+    @classmethod
+    def create(cls, kwargs):
+        """
+        Factory method. Will create a new PRecord of the current type and assign the values
+        specified in kwargs.
+        """
+        if isinstance(kwargs, cls):
+            return kwargs
+
+        return cls(**kwargs)
+
+    def __reduce__(self):
+        # Pickling support
+        return _restore_pickle, (self.__class__, dict(self),)
+
+    def serialize(self, format=None):
+        """
+        Serialize the current PRecord using custom serializer functions for fields where
+        such have been supplied.
+        """
+        def _serialize(k, v):
+            serializer = self.__class__._precord_fields[k].serializer
+            if isinstance(v, PRecord) and serializer is _PRECORD_NO_SERIALIZER:
+                return v.serialize(format)
+
+            return serializer(format, v)
+
+        return dict((k, _serialize(k, v)) for k, v in self.items())
+
+class _PRecordEvolver(PMap._Evolver):
+    __slots__ = ('_destination_cls', '_invariant_error_codes', '_missing_fields')
+
+    def __init__(self, cls, *args):
+        super(_PRecordEvolver, self).__init__(*args)
+        self._destination_cls = cls
+        self._invariant_error_codes = []
+        self._missing_fields = []
+
+    def __setitem__(self, key, original_value):
+        field = self._destination_cls._precord_fields.get(key)
+        if field:
+            try:
+                value = field.factory(original_value)
+            except InvariantException as e:
+                self._invariant_error_codes += e.invariant_errors
+                self._missing_fields += e.missing_fields
+                return
+
+            if field.type and not any(isinstance(value, t) for t in field.type):
+                raise TypeError("Invalid type for field {0}.{1}, was {2}".format(self._destination_cls.__name__, key, type(value)))
+
+            is_ok, error_code = field.invariant(value)
+            if not is_ok:
+                self._invariant_error_codes.append(error_code)
+
+            super(_PRecordEvolver, self).__setitem__(key, value)
+        else:
+            raise AttributeError("'{0}' is not among the specified fields for {1}".format(key, self._destination_cls.__name__))
+
+    def persistent(self):
+        cls = self._destination_cls
+        is_dirty = self.is_dirty()
+        pm = super(_PRecordEvolver, self).persistent()
+        if is_dirty or not isinstance(pm, cls):
+            result = cls(_precord_buckets=pm._buckets, _precord_size=pm._size)
+        else:
+            result = pm
+
+        if cls._precord_mandatory_fields:
+            self._missing_fields += tuple('{0}.{1}'.format(cls.__name__, f) for f
+                                          in (cls._precord_mandatory_fields - set(result.keys())))
+
+        if self._invariant_error_codes or self._missing_fields:
+            raise InvariantException(tuple(self._invariant_error_codes), tuple(self._missing_fields),
+                                     'Field invariant failed')
+
+        error_codes = tuple(error_code for is_ok, error_code in
+                            (invariant(result) for invariant in cls._precord_invariants) if not is_ok)
+        if error_codes:
+            raise InvariantException(error_codes, (), 'Global invariant failed')
+
+        return result
+
+
+############## Transform ##################
+
+def _discard(evolver, key):
+    try:
+        del evolver[key]
+    except KeyError:
+        pass
+
+# Transformations
+inc = lambda x: x + 1
+"""
+Add one to the current value
+"""
+
+dec = lambda x: x - 1
+"""
+Subtract one from the current value
+"""
+
+discard = _discard
+"""
+Discard the element and returns a structure without the discarded element.
+"""
+
+# Matchers
+def rex(expr):
+    """
+    Regular expression matcher to use together with transform functions
+    """
+    r = re.compile(expr)
+    return lambda key: isinstance(key, six.string_types) and r.match(key)
+
+
+ny = lambda _: True
+"""
+Matcher that matches any value
+"""
+
+# Support functions
+def _chunks(l, n):
+    for i in range(0, len(l), n):
+        yield l[i:i + n]
+
+
+def _transform(structure, transformations):
+    r = structure
+    for path, command in _chunks(transformations, 2):
+        r = _do_to_path(r, path, command)
+    return r
+
+
+def _do_to_path(structure, path, command):
+    if not path:
+        return command(structure) if callable(command) else command
+
+    kvs = _get_keys_and_values(structure, path[0])
+    return _update_structure(structure, kvs, path[1:], command)
+
+
+def _items(structure):
+    try:
+        return structure.items()
+    except AttributeError:
+        return list(enumerate(structure))
+
+
+def _get(structure, key, default):
+    try:
+        return structure[key]
+    except (IndexError, KeyError):
+        return default
+
+
+def _get_keys_and_values(structure, key_spec):
+    if callable(key_spec):
+        return [(k, v) for k, v in _items(structure) if key_spec(k)]
+
+    return [(key_spec, _get(structure, key_spec, pmap()))]
+
+
+def _update_structure(structure, kvs, path, command):
+    e = structure.evolver()
+    for k, v in kvs:
+        if not path and command is discard:
+            discard(e, k)
+        else:
+            result = _do_to_path(v, path, command)
+            if result is not v:
+                e[k] = result
+    return e.persistent()
